

from typing import Optional, List
import os
import traceback
import time

import pandas as pd

from ...util.mail_retriever import MailAttachmentRetriever, IMAP_SPType, UID_FILE_NAME
from ...util.wechat_bot import WechatBot
from ...constant import EMailParserType
from ..wrapper.mysql import DerivedDatabaseConnector
from ..view.derived_models import FOFNav  # , FOFUnconfirmedNav
from ..api.derived import DerivedDataApi
from .parser_runner import ParserRunner


class HedgeFundNAVReader:

    def __init__(self, manager_id: str, read_dir: str, user_name: str, password: str):
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)

        self._read_dir = read_dir
        assert os.path.isdir(self._read_dir), f'arg dump_dir should be a directory (now){self._read_dir}'

        self._manager_id = manager_id
        self._user_name = user_name
        self._password = password
        self._wechat_bot = WechatBot()

        self._parser_runner = ParserRunner(manager_id, EMailParserType.E_PARSER_NAV)

    def _notify_error_event(self, err_msg: str):
        print(f'[read_navs_and_dump_to_db] {err_msg}')
        self._wechat_bot.send_hedge_fund_nav_update_failed(err_msg)

    def read_navs_and_dump_to_db(self):
        try:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'rb') as f:
                uid_last = f.read()
                if not uid_last:
                    uid_last = None
        except Exception as e:
            self._notify_error_event(f'read uid file failed (e){e}, use None instead(read all emails)')
            uid_last = None

        try:
            mar = MailAttachmentRetriever(self._read_dir, ['xls', 'xlsx', 'pdf'])
            data = mar.get_excels(IMAP_SPType.IMAP_QQ, self._user_name, self._password, uid_last)
        except Exception as e:
            self._notify_error_event(f'FATAL ERROR!! get new data of hedge fund nav failed (e){e}')
            return

        uid_last_succeed: Optional[bytes] = None
        df_list: List[pd.DataFrame] = []
        for name, comp_date in data.items():
            uid, file_path = comp_date

            try:
                df = self._parser_runner.parse_file(file_path, name)
                if df is None:
                    raise NotImplementedError('unknown hedge fund/fof nav file from attachment')
            except Exception as e:
                self._notify_error_event(f'{e} (parse) (name){name} (file_path){file_path}')
                continue

            try:
                df = df.drop(columns=['fof_name', 'v_nav'], errors='ignore')
                df['manager_id'] = self._manager_id
                print(df)
                df = self._dump_to_db(df)
                if df is not None:
                    df_list.append(df)
                else:
                    print(f'[read_navs_and_dump_to_db] duplicated data, do not process it (name){name}')
                # 走到这里都认为是已经处理完了这条数据
                uid_last_succeed = uid
                time.sleep(1)
            except Exception as e:
                traceback.print_exc()
                self._notify_error_event(f'{e} (dump) (name){name} (file_path){file_path}')
                break

        if df_list:
            try:
                whole_df = pd.concat(df_list).set_index('fof_id')
                print(whole_df)
            except Exception as e:
                self._notify_error_event(f'{e} (concat)')
                whole_df = None
            else:
                # self._wechat_bot.send_hedge_fund_nav_update(whole_df)
                print(f'[read_navs_and_dump_to_db] done (uid_last){uid_last_succeed} (df){whole_df}')
        else:
            whole_df = None
            print(f'[read_navs_and_dump_to_db] no new data this time, done (uid_last){uid_last_succeed}')
        # 记录下成功的最后一个uid
        if whole_df is not None and uid_last_succeed is not None:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'wb') as f:
                f.write(uid_last_succeed)
            return whole_df
        return

    def _dump_to_db(self, df: pd.DataFrame):
        def _check_after_merged(x: pd.Series, now_df: pd.DataFrame):
            try:
                now_data = now_df[(now_df.fof_id == x.fof_id) & (now_df.datetime == x.datetime)]
                if now_data[['nav', 'acc_net_value']].iloc[0].astype('float64').equals(x[['nav', 'acc_net_value']].astype('float64')):
                    return pd.Series(dtype='object')
                else:
                    return x
            except (KeyError, IndexError):
                return x

        assert df.datetime.nunique() == 1, 'should have single datetime'
        # now_df = DerivedDataApi().get_fof_nav_unconfirmed(manager_id=self._manager_id, fof_id_list=list(df.fof_id.unique()))
        now_df = DerivedDataApi().get_fof_nav(manager_id=self._manager_id, fof_id_list=list(df.fof_id.unique()))
        if now_df is not None and not now_df.empty:
            # 同产品同日期的净值如果已经存在了且没有变化 就不写DB了
            now_df = now_df.drop(columns=['update_time', 'create_time', 'is_deleted', 'adjusted_nav', 'ta_factor', 'volume', 'mv', 'ret']).sort_values(by=['manager_id', 'fof_id', 'datetime']).drop_duplicates(subset=['manager_id', 'fof_id', 'datetime'], keep='last')
            now_df = now_df.astype({'nav': 'float64', 'acc_net_value': 'float64'})
            df = df.reindex(columns=now_df.columns).astype(now_df.dtypes.to_dict())
            df = df.merge(now_df, how='left', on=['manager_id', 'fof_id', 'datetime', 'nav', 'acc_net_value'], indicator=True, validate='one_to_one')
            df = df[df._merge == 'left_only'].drop(columns=['_merge'])
            if df.empty:
                return
            # FIXME 没想到特别好的方法 遍历每一行再check一下
            df['datetime'] = pd.to_datetime(df.datetime, infer_datetime_format=True).dt.date
            df = df.apply(_check_after_merged, axis=1, now_df=now_df)
            if df.empty:
                return
            df = df.set_index(['manager_id', 'fof_id', 'datetime'])
            print(df)
            df.update(now_df.set_index(['manager_id', 'fof_id', 'datetime']), overwrite=False)
            df = df.reset_index()
            df['datetime'] = df.datetime.map(lambda x: x.date())
            print(df)
            # 先删后添
            for fof_id in df.fof_id.unique():
                DerivedDataApi().delete_fof_nav(manager_id=self._manager_id, fof_id=fof_id, date_list=df[df.fof_id == fof_id].datetime.to_list())
                # DerivedDataApi().delete_fof_nav_unconfirmed(manager_id=self._manager_id, fof_id=fof_id, date_list=df[df.fof_id == fof_id].datetime.to_list())
        df.to_sql(FOFNav.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        # df.to_sql(FOFUnconfirmedNav.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        return df


if __name__ == '__main__':
    try:
        email_data_dir = os.environ['SURFING_EMAIL_DATA_DIR']
        user_name = os.environ['SURFING_EMAIL_USER_NAME']
        password = os.environ['SURFING_EMAIL_PASSWORD']
    except KeyError as e:
        import sys
        sys.exit(f'can not found enough params in env (e){e}')

    MANAGER_ID = 'py1'

    hf_nav_r = HedgeFundNAVReader(MANAGER_ID, email_data_dir, user_name, password)
    hf_nav_r.read_navs_and_dump_to_db()



import pandas as pd
import numpy as np
import datetime 
import matplotlib as mpl
import matplotlib.pyplot as plt
import statsmodels.api as sm
mpl.rcParams['font.family'] = ['Heiti TC']


class TrendCalculator:
    
    @staticmethod
    # 短线下穿长线
    def cross_down(index_price, short_term_ma, long_term_ma):
        index_price['_tmp'] = index_price[short_term_ma] < index_price[long_term_ma]
        return index_price['_tmp'] & (~ index_price['_tmp'].shift(1).fillna(False)) 
    
    @staticmethod
    # 两条线过去一端时间没有相交，且line_1在line_2上面
    def up_with_no_cross(index_price, line_1, line_2, term):
        index_price['_tmp'] = (index_price[line_1] > index_price[line_2]) * 1
        return index_price['_tmp'].rolling(term).min() > 0

    @staticmethod
    # 短线下穿长线
    def cross_up(index_price, short_term_ma, long_term_ma):
        index_price['_tmp'] = index_price[short_term_ma] > index_price[long_term_ma]
        return index_price['_tmp'] & (~ index_price['_tmp'].shift(1).fillna(False)) 

    @staticmethod
    # 两线幅度差
    def diff_rate(index_price, line_1, line_2):
        return index_price[line_1] / index_price[line_2] - 1

class TrendStrategy:

    DATA_SOURCE = '/Users/huangkejia/Downloads/test/sp500rmb.xlsx'
    YEAR = 242
    RISK_FEE_RATE = 0.025

    def __init__(self):
        pass

    def init(self, df=None):
        dic = {
            '代码':'index_id',
            '日期':'datetime',
            '开盘价(元)':'open',
            '最高价(元)':'high',
            '最低价(元)':'low',
            '收盘价(元)':'close',
        }
        if df is None:
            self.index_price = pd.read_excel(self.DATA_SOURCE).iloc[:-2]
            self.index_price = self.index_price.rename(columns = dic).drop(['名称','成交额(百万)','成交量(股)'], axis=1)
            self.index_price['index_id'] = 'sp500rmb'
            self.index_price = self.index_price.set_index('datetime')
            self.index_price.index = pd.to_datetime(self.index_price.index).date
        else:
            self.index_price = df
            self.index_price = self.index_price.set_index('datetime')
        self.dts = self.index_price.index.tolist()
        self.index_id = self.index_price.index_id.unique().tolist()[0]

    def fac_calculator(self):
        self.index_price.index = pd.to_datetime(self.index_price.index).date
        self.index_price.loc[:,'ret'] = self.index_price['close'].pct_change(1)
        self.index_price.loc[:,'std5'] = self.index_price['ret'].rolling(window=5).std(ddof=1)
        self.index_price.loc[:,'std10'] = self.index_price['ret'].rolling(window=10).std(ddof=1)
        self.index_price.loc[:,'std120'] = self.index_price['ret'].rolling(window=120).std(ddof=1)   

        term = 120
        self.index_price['term_high'] = self.index_price['high'].rolling(window=term,min_periods=1).max()
        self.index_price['term_low'] = self.index_price['low'].rolling(window=term,min_periods=1).min()
        self.index_price['rsi'] = (self.index_price['close'] - self.index_price['term_low']) / (self.index_price['term_high'] - self.index_price['term_low']) * 100
        self.index_price['rsi_ma_x1'] = self.index_price['rsi'].rolling(window=5, min_periods=1).mean()
        self.index_price['rsi_ma_x2'] = self.index_price['rsi_ma_x1'].rolling(window=3, min_periods=1).mean()
        self.index_price['rsi_ma_x3'] = self.index_price['rsi_ma_x2'].rolling(window=2, min_periods=1).mean()

        short_term = 10
        self.index_price['term_s_high'] = self.index_price['high'].rolling(window=short_term,min_periods=1).max()
        self.index_price['term_s_low'] = self.index_price['low'].rolling(window=short_term,min_periods=1).min()
        self.index_price['rsi_s'] = (self.index_price['close'] - self.index_price['term_s_low']) / (self.index_price['term_s_high'] - self.index_price['term_s_low']) * 100
        self.index_price['rsi_s_ma_x1'] = self.index_price['rsi_s'].rolling(window=5, min_periods=1).mean()
        self.index_price['rsi_s_ma_x2'] = self.index_price['rsi_s_ma_x1'].rolling(window=3, min_periods=1).mean()
        self.index_price['rsi_s_ma_x3'] = self.index_price['rsi_s_ma_x2'].rolling(window=2, min_periods=1).mean()

        self.index_price['ma_2d'] = self.index_price['close'].rolling(2).mean()
        self.index_price['ma_1w'] = self.index_price['close'].rolling(5).mean()
        self.index_price['ma_2w'] = self.index_price['close'].rolling(10).mean()
        self.index_price['ma_3w'] = self.index_price['close'].rolling(15).mean()
        self.index_price['ma_1m'] = self.index_price['close'].rolling(20).mean()
        self.index_price['ma_3m'] = self.index_price['close'].rolling(60).mean()
        self.index_price['ma_6m'] = self.index_price['close'].rolling(121).mean()
        self.index_price['ma_9m'] = self.index_price['close'].rolling(180).mean()
        self.index_price['ma_1y'] = self.index_price['close'].rolling(self.YEAR).mean()
        self.index_price['ma_2y'] = self.index_price['close'].rolling(self.YEAR*2).mean()
        self.index_price['ma_3y'] = self.index_price['close'].rolling(self.YEAR*3).mean()
        self.index_price['ma_5y'] = self.index_price['close'].rolling(self.YEAR*5).mean()
        self.index_price['ma_7y'] = self.index_price['close'].rolling(self.YEAR*7).mean()
        self.index_price['y1_diff'] = self.index_price['close'] < self.index_price['ma_1y']
        self.index_price['m6_diff'] = self.index_price['close'] < self.index_price['ma_6m']
        self.index_price['m3_diff'] = self.index_price['close'] < self.index_price['ma_3m']
        self.index_price['m1_diff'] = self.index_price['close'] < self.index_price['ma_1m']
        self.index_price['d20_rise'] = self.index_price['ma_1m'] > self.index_price['ma_1m'].shift(1)
        self.index_price['d15_rise'] = self.index_price['ma_3w'] > self.index_price['ma_3w'].shift(1)
        self.index_price['d10_rise'] = self.index_price['ma_2w'] > self.index_price['ma_2w'].shift(1)
        self.index_price['d5_rise'] = self.index_price['ma_1w'] > self.index_price['ma_1w'].shift(1)
        self.index_price['long_term_rise'] = (self.index_price['ma_5y'] > self.index_price['ma_7y']) & \
                                        (self.index_price['ma_1y'] > self.index_price['ma_2y']) & \
                                        (self.index_price['ma_1m'] > self.index_price['ma_6m'])

        self.index_price['close_y1_diff'] = TrendCalculator.diff_rate(self.index_price, 'close','ma_1y')
        self.index_price['close_9m_diff'] = TrendCalculator.diff_rate(self.index_price, 'close','ma_9m')
        self.index_price['close_6m_diff'] = TrendCalculator.diff_rate(self.index_price, 'close','ma_6m')
        self.index_price['close_3m_diff'] = TrendCalculator.diff_rate(self.index_price, 'close','ma_3m')
        self.index_price['close_1m_diff'] = TrendCalculator.diff_rate(self.index_price, 'close','ma_1m')
        self.index_price['1m_y1_diff'] = TrendCalculator.diff_rate(self.index_price, 'ma_1m','ma_1y')
        self.index_price['1w_y1_diff'] = TrendCalculator.diff_rate(self.index_price, 'ma_1w','ma_1y')
        self.index_price['ma_y1_std242'] = self.index_price['ma_1y'].pct_change(1).rolling(242).std(ddof=1)
        self.index_price['ma_y1_std120'] = self.index_price['ma_1y'].pct_change(1).rolling(120).std(ddof=1)
        self.index_price['ma_y1_std60'] = self.index_price['ma_1y'].pct_change(1).rolling(60).std(ddof=1)
        self.index_price['ma_y1_std30'] = self.index_price['ma_1y'].pct_change(1).rolling(30).std(ddof=1)
        self.index_price['ma_y1_std20'] = self.index_price['ma_1y'].pct_change(1).rolling(20).std(ddof=1)
        self.index_price['ma_y1_std10'] = self.index_price['ma_1y'].pct_change(1).rolling(10).std(ddof=1)
        self.index_price['ma_y1_std5'] = self.index_price['ma_1y'].pct_change(1).rolling(5).std(ddof=1)
        self.index_price['top_fac'] = TrendCalculator.cross_down(self.index_price,'close','ma_9m') & \
                                (TrendCalculator.up_with_no_cross(self.index_price, 'close','ma_9m', 5).shift(1))#20
        self.index_price['bottom_fac'] = (self.index_price['rsi_ma_x3'].rolling(5).min() < 10)# & \
        self.index_price['stop_earn_fac'] = TrendCalculator.cross_up(self.index_price, 'ma_5y', 'ma_7y')


    def make_trade_signal(self, df:pd.DataFrame, start_fac:str='top_fac', end_fac:str='bottom_fac', if_stop_loss:bool=False):
        trade_status = False
        open_price = None
        start_signal_list = []
        end_signal_list = []
        for _dt in df.index:
            if not trade_status:
                if df.loc[_dt, start_fac]:
                    start_signal_list.append(_dt)
                    open_price = df.loc[_dt,'close']
                    trade_status = True
            if trade_status:
                if df.loc[_dt, end_fac]:
                    end_signal_list.append(_dt)
                    trade_status = False
                    continue
                if if_stop_loss:
                    if df.loc[_dt,'close'] / open_price > 1.35:
                        end_signal_list.append(_dt)
                        trade_status = False
        if trade_status:
            end_signal_list.append(_dt)
        return start_signal_list, end_signal_list

    def plot_signal_point(self, start_signal_list, end_signal_list, line_1, line_2, is_plot_type_1=False, is_plot_type_2=False):
        each_invest_captical = 100
        last_idx = 0
        date_list = [self.dts[0]]
        year_length = 5
        while last_idx < len(start_signal_list):
            d0 = start_signal_list[last_idx]
            d1 = end_signal_list[last_idx]
            while ((d1 - d0).days < 365 * year_length) and (last_idx < len(start_signal_list)):
                last_idx += 1
                if last_idx < len(start_signal_list):
                    d1 = end_signal_list[last_idx]
                else:
                    break
            last_idx += 1
            date_list.append(d0 - datetime.timedelta(1))
            date_list.append(d1 + datetime.timedelta(1))
        df_res = []
        for idx, _dt in enumerate(date_list[:-1]):
            dt =  date_list[date_list.index(_dt) + 1]
            _dt_ = _dt - datetime.timedelta(days=1)
            dt_ = dt + datetime.timedelta(days=1)
            start_signal_dts = [dt_i for dt_i in start_signal_list if (dt_i >= _dt) and (dt_i <= dt)]
            end_signal_dts = [dt_i for dt_i in end_signal_list if (dt_i >= _dt) and (dt_i <= dt)]   
            res = []
            for dt_1, dt_2 in zip(start_signal_dts,end_signal_dts):
                start_price = self.index_price.loc[dt_1,'close']
                end_price = self.index_price.loc[dt_2,'close']
                ret = (start_price - end_price) / start_price
                dic = {
                    'start_date':dt_1,
                    'end_date':dt_2,
                    'start_price':start_price,
                    'end_price':end_price,
                    'ret':ret,
                    'win':ret>0,
                    'amount': (ret+1)*each_invest_captical
                }
                res.append(dic)
            _df_i = pd.DataFrame(res)
            red_start_date = [i.start_date for i in _df_i.itertuples() if i.ret >0 ]
            green_start_date = [i.start_date for i in _df_i.itertuples() if i.ret <= 0]
            red_end_date = [i.end_date for i in _df_i.itertuples() if i.ret >0 ]
            green_end_date = [i.end_date for i in _df_i.itertuples() if i.ret <= 0]
            df_res.append(_df_i)
            if is_plot_type_1:
                _index_price = self.index_price[['close',line_1,line_2]].loc[_dt_:dt_].copy()
                _index_price[['close']].plot.line(figsize=(16,8), fontsize=15,linewidth=2, zorder=1)
                success_c = 'red'
                fail_c = 'limegreen'
                size = 250
                up_loc = 1.03
                low_loc = 0.97
                plt.scatter(x=red_start_date,y=_index_price[['close']].loc[red_start_date] * up_loc, c=success_c, marker='$1$', s = size, zorder=5)
                plt.scatter(x=green_start_date,y=_index_price[['close']].loc[green_start_date] * low_loc, c=fail_c, marker='$1$', s = size, zorder=5)
                plt.scatter(x=red_end_date,y=_index_price[['close']].loc[red_end_date] * up_loc, c=success_c, marker='$2$', s = size, zorder=5)
                plt.scatter(x=green_end_date,y=_index_price[['close']].loc[green_end_date] * low_loc, c=fail_c, marker='$2$', s = size, zorder=5)
                plt.plot(_index_price.index, _index_price[line_1], c='gold',linewidth=1.5, zorder=2, label=line_1)
                plt.plot(_index_price.index, _index_price[line_2], c='lightgray',linewidth=1.5, zorder=3,label=line_2)
                try:
                    _win_rate = round(_df_i.win.sum() / _df_i.shape[0] * 100,1)
                    plt.title(f'{self.index_id} {dt.year} top win rate {_win_rate}%', fontsize=20)
                except:
                    plt.title(f'{self.index_id} {dt.year}', fontsize=20)
                plt.legend(fontsize=15)
                plt.grid()
                plt.show()
            elif is_plot_type_2:
                _line = 't_stats'
                _index_price = self.index_price[['close',line_2,_line]].loc[_dt_:dt_].copy()
                success_c = 'red'
                fail_c = 'limegreen'
                size = 250
                up_loc = 1.03
                low_loc = 0.97
                fig, ax1 = plt.subplots(figsize=(16,8))
                ax1.set_xlabel('date', fontsize=15)
                ax1.set_ylabel('price', fontsize=15)
                ax1.plot(_index_price.index, _index_price.close, label='close')
                ax1.plot(_index_price.index, _index_price[line_2], label=line_2)
                plt.scatter(x=red_start_date,y=_index_price[['close']].loc[red_start_date] * up_loc, c=success_c, marker='$1$', s = size, zorder=5)
                plt.scatter(x=green_start_date,y=_index_price[['close']].loc[green_start_date] * low_loc, c=fail_c, marker='$1$', s = size, zorder=5)
                plt.scatter(x=red_end_date,y=_index_price[['close']].loc[red_end_date] * up_loc, c=success_c, marker='$2$', s = size, zorder=5)
                plt.scatter(x=green_end_date,y=_index_price[['close']].loc[green_end_date] * low_loc, c=fail_c, marker='$2$', s = size, zorder=5)
                plt.yticks(fontsize=14)
                plt.xticks(fontsize=14)
                plt.legend(fontsize=18)
                ax2 = ax1.twinx() 
                color = 'gold'
                ax2.set_ylabel(_line,fontsize=15) 
                ax2.plot(_index_price.index, _index_price[_line], label=_line, color = color)
                try:
                    _win_rate = round(_df_i.win.sum() / _df_i.shape[0] * 100,1)
                    plt.title(f'{self.index_id} {dt.year} top win rate {_win_rate}%', fontsize=20)
                except:
                    plt.title(f'{self.index_id} {dt.year}', fontsize=20)
                plt.yticks(fontsize=14)
                plt.legend(fontsize=18)
                plt.grid()
                plt.show()
        df_total = pd.concat(df_res, axis=0)
        return df_total

    def consecutive_analysis(self, df, direction='loss'):
        res_list = []
        tmp_list = []
        for r in df.itertuples():
            if direction == 'loss':
                if r.ret <= 0:
                    tmp_list.append({
                        'start_date':r.start_date,
                        'end_date':r.end_date,
                        'ret':r.ret,
                    })
                if r.ret > 0 and len(tmp_list) > 0:
                    res_list.append(tmp_list)
                    tmp_list = []
            else:
                if r.ret >= 0:
                    tmp_list.append({
                        'start_date':r.start_date,
                        'end_date':r.end_date,
                        'ret':r.ret,
                    })
                if r.ret < 0 and len(tmp_list) > 0:
                    res_list.append(tmp_list)
                    tmp_list = []

        consecutive_numbers = max([len(i) for i in res_list])
        length_dic = []
        _res_list = []
        for i in res_list:
            dic = {}
            dic['start_date'] = i[0]['start_date']
            dic['end_date'] = i[-1]['end_date']
            dic['day_length'] = (dic['end_date'] - dic['start_date']).days
            dic['max_diff'] = sum([_i['ret'] for _i in i])
            length_dic.append(dic)
            _res_list.append(dic['max_diff'])
        if direction == 'loss':
            result = [i for i in length_dic if i['max_diff'] == min(_res_list)][0]
        else:
            result = [i for i in length_dic if i['max_diff'] == max(_res_list)][0]
        result['consecutive_numbers'] = consecutive_numbers
        return result

    def asset_constancy_analysis(self):
        index_close = self.index_price[['close']]
        index_close = index_close.reset_index().rename(columns={'index':'datetime'})
        index_close['year_month'] = index_close.datetime.map(lambda x: f'{x.year}_{x.month}' )
        index_close = index_close.set_index('year_month')
        index_list = sorted(index_close.index.unique().tolist())
        res = []
        for index_i in index_list:
            _df = index_close.loc[index_i]
            if _df.shape[0] < 5:
                continue
            else:
                trade_year = _df.shape[0] / self.YEAR
                annualized_ret = np.exp(np.log(_df['close'].values[-1]/_df['close'].values[0])/trade_year ) - 1
                annualized_vol = (_df.close/_df.close.shift(1)).std(ddof=1) * np.sqrt(self.YEAR)
                sharpe = (annualized_ret - self.RISK_FEE_RATE) / annualized_vol
                dic = {'datetime':index_i,'sharpe':sharpe}
                res.append(dic)
        sharpe_result = pd.DataFrame(res)
        result = []
        for lag_m in range(1,61):
            y = sharpe_result.sharpe.shift(-lag_m).dropna()
            x = sharpe_result.loc[:y.shape[0]-1].sharpe
            est = sm.OLS(y, sm.add_constant(x, prepend=False)).fit()
            p = est.params.sharpe
            result.append({'lag':lag_m,'显著性':p})
        pd.DataFrame(result).set_index('lag').plot.bar(figsize=(16,8), fontsize=15, rot=0)
        plt.title(f'{self.index_id}业绩持续性',fontsize=25)
        plt.legend(fontsize=20)
        plt.xlabel('lag', fontsize=18)
        plt.show()
    
    def get_backtest_result(self, df):
        year_lens = len(list(set([i.year for i in self.dts])))
        date_list = [(r.end_date - r.start_date).days for r in df.itertuples()] 
        stats_res = {}
        stats_res['asset'] = self.index_id
        stats_res['start_date'] = self.index_price.index.tolist()[0]
        stats_res['end_date'] = self.index_price.index.tolist()[-1]
        stats_res['trade_num'] = df.shape[0]
        stats_res['win_trade_num'] = sum(df['win'] * 1)
        stats_res['loss_trade_num'] = df.shape[0] - sum(df['win'] * 1)
        stats_res['stop_earning_num'] = sum(df['ret'] < -0.15)
        stats_res['averager_holding_periods'] = int(np.mean(date_list))
        stats_res['max_holding_periods'] = int(np.max(date_list))
        stats_res['min_holding_periods'] = int(np.min(date_list))
        stats_res['max_earn_single_trade'] = df['ret'].max() 
        stats_res['max_loss_single_trade'] = df['ret'].min() 
        stats_res['earn_sum'] = df[df['ret'] > 0].ret.sum()
        stats_res['loss_sum'] = df[df['ret'] < 0].ret.sum()
        stats_res['earn_loss_rate'] = - stats_res['earn_sum'] / stats_res['loss_sum'] 
        stats_res['total_win_rate'] = stats_res['win_trade_num'] / (stats_res['win_trade_num'] + stats_res['loss_trade_num'])
        stats_res['num_per_year'] = stats_res['trade_num'] / year_lens
        stats_res['total_ret'] = df.amount.sum() / (df.shape[0] * 100) - 1
        stats_res['ret_per_trade'] = stats_res['total_ret'] / df.shape[0]
        lose_result = self.consecutive_analysis(df, direction='loss')
        stats_res['consecutive_loss'] = lose_result['consecutive_numbers']
        stats_res['consecutive_loss_start_date'] = lose_result['start_date']
        stats_res['consecutive_loss_end_date'] = lose_result['end_date']
        stats_res['max_consecutive_loss'] = lose_result['max_diff']
        win_result = self.consecutive_analysis(df, direction='win')
        stats_res['consecutive_win'] = win_result['consecutive_numbers']
        stats_res['consecutive_win_start_date'] = win_result['start_date']
        stats_res['consecutive_win_end_date'] = win_result['end_date']
        stats_res['max_consecutive_earn'] = win_result['max_diff']

        int_list = ['trade_num',
                    'win_trade_num',
                    'loss_trade_num',
                    'averager_holding_periods',
                    'max_holding_periods',
                    'min_holding_periods',
                    'stop_earning_num',
                    'consecutive_loss',
                    'consecutive_win']

        rate_list = ['earn_loss_rate',
                    'total_win_rate',
                    'total_ret',
                    'max_earn_single_trade',
                    'max_loss_single_trade',
                    'max_consecutive_loss',
                    'max_consecutive_earn']

        float_list = ['earn_sum','loss_sum','num_per_year']

        bp_list = ['ret_per_trade']

        for i in int_list:
            stats_res[i] = str(int(stats_res[i]))
        for i in rate_list:
            stats_res[i] = str(round((stats_res[i] * 100), 2)) + '%'

        for i in float_list:
            stats_res[i] = str(round((stats_res[i]), 3)) 

        for i in bp_list:
            stats_res[i] = str(round((stats_res[i] * 10000), 1)) +'bp'
        
        dic = {
            'asset':'资产', 
            'start_date':'回测开始日', 
            'end_date':'回测结束日', 
            'trade_num':'交易次数', 
            'win_trade_num':'盈利交易次数',
            'loss_trade_num':'亏损交易次数', 
            'stop_earning_num':'止损交易次数', 
            'averager_holding_periods':'平均开仓周期', 
            'num_per_year':'平均年开仓次数',
            'max_earn_single_trade':'最大单笔盈利',
            'max_loss_single_trade':'最大单笔亏损', 
            'max_holding_periods':'最大开仓天数', 
            'min_holding_periods':'最低开仓天数',
            'earn_sum':'盈利总额', 
            'loss_sum':'亏损总额', 
            'earn_loss_rate':'盈亏比',
            'total_win_rate':'总胜率', 
            'total_ret':'总收益', 
            'ret_per_trade':'每单平均收益',
            'consecutive_loss':'最大连续亏损交易次数',
            'consecutive_win':'最大连续盈利交易次数',
            'consecutive_loss_start_date':'最大连续亏损开始日',
            'consecutive_loss_end_date':'最大连续亏损结束日',
            'consecutive_win_start_date':'最大连续盈利开始日',
            'consecutive_win_end_date':'最大连续盈利结束日',
            'max_consecutive_loss':'最大连续亏损占单次开仓比例',
            'max_consecutive_earn':'最大连续盈利占单次开仓比例',
        }
        result = pd.DataFrame([stats_res], index=['回测记录']).rename(columns = dic)
        result = result.drop(['盈利总额','亏损总额','盈亏比'], axis=1)
        
        return result.T


if __name__ == "__main__":
    '''
    self = TrendStrategy()
    self.init()
    self.asset_constancy_analysis()

    strategy 1  未优化
    标普500过去12个月的价格均线和月底价格作比较，每个月根据信号来测试。
    如果月底标普500的价格高于过去12个月的均价（包含当月），则持有标普；否则卖出标普
    (月度计算)
    df = self.index_price.copy()
    df.index = pd.to_datetime(df.index)
    df = df[['close']].resample('M').last().copy()
    df.index = pd.Series(df.index).dt.date
    df = df.reset_index().copy()
    index_list = self.index_price.index.values
    real_index = []
    for r in df.itertuples():
        real_index.append(index_list[index_list <= r.index].tolist()[-1])
    df.index = real_index
    df['ma_1y'] = df[['close']].rolling(window=12).mean()
    df['top_fac'] = TrendCalculator.cross_down(df,'close','ma_1y')
    df['bottom_fac'] = TrendCalculator.cross_up(df,'close','ma_1y')

    start_signal_list, end_signal_list = self.make_trade_signal(df=df, start_fac='top_fac', end_fac='bottom_fac')
    df = self.plot_signal_point(start_signal_list, end_signal_list, 'ma_1w', 'ma_1y',is_plot_type_1=True)
    result = self.get_backtest_result(df)
    
    
    strategy 2  未优化
    标普500过去12个月的总回报。如果标普过去12个月的总回报大于0，则继续持有，否则卖出 (月度计算)
    
    df = self.index_price.copy()
    df.index = pd.to_datetime(df.index)
    df = df[['close']].resample('M').last().copy()
    df.index = pd.Series(df.index).dt.date
    df = df.reset_index().copy()
    index_list = self.index_price.index.values
    real_index = []
    for r in df.itertuples():
        real_index.append(index_list[index_list <= r.index].tolist()[-1])
    df.index = real_index
    df['top_fac'] = df['close'].pct_change(12) > 0
    df['bottom_fac'] = df['close'].pct_change(12) < 0
    start_signal_list, end_signal_list = self.make_trade_signal(df=self.index_price, start_fac='top_fac', end_fac='bottom_fac')
    df = self.plot_signal_point(start_signal_list, end_signal_list, 'ma_1w', 'ma_1y', is_plot_type_1=True)
    result = self.get_backtest_result(df)


    strategy 3 未优化
    标普500过去12个月的总回报。如果标普过去12个月的总回报大于0，则继续持有，否则卖出 (日度计算)|
    self.index_price['top_fac'] = self.index_price['close'].pct_change(self.YEAR) > 0
    self.index_price['bottom_fac'] = self.index_price['close'].pct_change(self.YEAR) < 0
    start_signal_list, end_signal_list = self.make_trade_signal(df=self.index_price, start_fac='top_fac', end_fac='bottom_fac')
    df = self.plot_signal_point(start_signal_list, end_signal_list, 'ma_1w', 'ma_1y',is_plot_type_1=True)
    result = self.get_backtest_result(df)

    strategy 4 增加因子做筛选 微调
    计算其过去12个月的日收益率的 Newey-West t统计量 日收益对时序做回归
    开仓条件:
        t统计量小于-0.8，收盘价 低于一年均线 0.02，6个月rsi 小于 35
    平仓条件
        t统计量大于-0.2

    t_stats = []
    for i in self.index_price.index.tolist():
        _df = self.index_price.loc[:i,'ret'].dropna()
        if _df.shape[0] < 242:
            t_stats.append(0)
        else:
            F_tau = _df.values.tolist()[-self.YEAR:]
            tau = [i + 1 for i in range(self.YEAR)]
            est = sm.OLS(F_tau, sm.add_constant(tau, prepend=False)).fit()
            t_stats.append(est.tvalues[0])

    self.index_price['t_stats'] = t_stats
    t_start  = -0.8
    t_end = -0.25
    m_y1_diff = -0.02
    rsi = 35
    
    self.index_price['top_fac'] = (self.index_price['t_stats'] < t_start) & \
                              (self.index_price['1m_y1_diff'] < m_y1_diff) & \
                              (self.index_price['rsi'] < rsi)
                             
    self.index_price['bottom_fac'] = self.index_price['t_stats'] > t_end
    start_signal_list, end_signal_list = self.make_trade_signal(df=self.index_price, start_fac='top_fac', end_fac='bottom_fac')
    df = self.plot_signal_point(start_signal_list, end_signal_list, 'ma_1w', 'ma_1y', is_plot_type_1=True, is_plot_type_2=False)
    test = self.get_backtest_result(df)
    '''



from .util.config import Configurator
from .util.struct import WxRawData, WxSeqItem, WxMsgItem
from .message_processor import IfaProcessor
from .message_receiver.wechat.message_receiver import MessageReceiver
import signal
import time, json
import threading
import sys
import libwxwork

class IfaRunner(object):

    def __init__(self, seq=None):
        self.seq = seq or 0
        
        self.processors = []

        self.received_signal = None
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)

    def register(self, processor: IfaProcessor):
        self.processors.append(processor)

    def process_monitor(self):
        corp_id = Configurator().config['wxwork']['corp_id']
        secret_key = Configurator().config['wxwork']['secret_key']
        private_key = Configurator().config['wxwork']['private_key']
        with open(private_key, 'r') as fin:
            private_key_content = fin.read()
        monitor = libwxwork.WxworkWrapper(corp_id, secret_key, private_key_content, -1)
        monitor.init()

        while not self.is_stopped():
            time.sleep(1)
            try:
                res = monitor.get(self.seq, 500)
                data = WxRawData(**json.loads(res))
                for item in data.chatdata:
                    decrypted = None
                    try:
                        decrypted_str = monitor.decode(item.encrypt_random_key, item.encrypt_chat_msg)
                        decrypted = json.loads(decrypted_str)
                    except:
                        print('!' * 10)
                        print(f'error: {decrypted_str}')
                        print('!' * 10)
                    if decrypted:
                        msg = WxMsgItem(decrypted)
                    self.process(msg)
            except Exception as e:
                print(e)
            self.seq = data.max_seq or self.seq

    def process_received(self):
        mr = MessageReceiver()
        t = threading.Thread(target=mr.run)
        t.setDaemon(True)
        t.start()

        while not self.is_stopped():
            time.sleep(1)
            try:
                data = mr.get_msg(500)
                for item in data:
                    self.process(item)
            except Exception as e:
                print(e)

        mr.stop()

    def process(self, msg: WxMsgItem):
        for pr in self.processors:
            pr.process(msg)

    def run(self):
        from .message_processor.wechat.message_processor import WechatMessageProcessor
        self.register(WechatMessageProcessor())

        ts = [
            # threading.Thread(target=self.process_monitor),
            threading.Thread(target=self.process_received)
        ]
        for t in ts:
            t.setDaemon(True)
            t.start()

        print('IFA service started.')

        while True:
            time.sleep(0.01)
            if self.is_stopped():
                print('Wxwork service exited. (sig){}'.format(self.received_signal))
                break

    def signal_handler(self, signum=None, frame=None):
        if self.received_signal is None:
            self.stop(signum=signum)

    def stop(self, signum=signal.SIGTERM):
        self.received_signal = signum

    def is_stopped(self):
        return self.received_signal is not None
